import csv
import numpy as np

class TimeSeries:
    def __init__(self, time_stamps, values):
        self.time_stamps = time_stamps
        self.values = values


    @classmethod
    def load(self, path, format):
        """
        Loads time series data from a file
        :param path: String. Path to the file that contains time series
        :param format: String. What is the format of the file being loaded from
        :return: A matrix that contains the time series data (numpy array form)
        """
        if format == "UCR":
            return TimeSeries.ucr_load(path)
        elif format == "Single Time Series":
            return TimeSeries.load_single_time_series(path)

    @classmethod
    def load_single_time_series(self, path):
        """
        This is for reading a single time series where every line in the file
        contains first the time stamp then the value
        :param path: path to the file
        :return: TimeSeries object
        """
        with open(path) as csv_file:
            lines = csv.reader(csv_file, delimiter = ',')
            time_stamps = []
            values = []
            for line in lines:
                try:
                    time_stamps.append(int(line[0]))
                    values.append(int(line[1]))
                except ValueError:
                    raise ValueError("There is a problem with the csv file format. Try changing the file encoding to Unicode without BOM.")

            return TimeSeries(time_stamps, values)

    @classmethod
    def ucr_load(self, path, with_time_stamps = False):
        """
        Loads from a csv file in UCR format. Every row contains a time series without timestamps. The first column of every
        row is the class time series belongs to.
        :param path: The file path
        :param with_time_stamps: If this is true then the file starts with explicit time stamps in its first row
        :return: numpy matrix that contains the loaded time series in its rows, an np array of labels of these time series.
        """
        if with_time_stamps == True:
            raise NotImplemented
        labels = []

        with open(path) as csv_file:
            lines = csv.reader(csv_file, delimiter=',')
            ls = []
            for line in lines:
                labels.append(int(line[0]))
                line.pop(0)
                try:
                    ls.append([float(x) for x in line])
                except ValueError:
                    raise ValueError("There is a problem with the csv file format. Try changing the file encoding to Unicode without BOM.")
        return np.array(ls), np.array(labels)







from scipy import stats
import GRAIL.SINK as SINK
import numpy as np


class CorrelationProtocol:
    def __init__(self, x, y, **kwargs):
        self.x = x
        self.y = y
        self.kwargs = kwargs

    def execute(self):
        raise NotImplemented


class Pearson(CorrelationProtocol):
    def __init__(self, x, y, **kwargs):
        super().__init__(x, y, **kwargs)
        self.similarity = True

    def execute(self):
        return stats.pearsonr(self.x, self.y)[0]


class ED(CorrelationProtocol):

    def __init__(self, x, y, **kwargs):
        super().__init__(x, y, **kwargs)
        self.similarity = False

    def execute(self):
        return np.linalg.norm(self.x - self.y)


class NCC(CorrelationProtocol):
    def __init__(self, x, y, **kwargs):
        super().__init__(x, y, **kwargs)
        self.similarity = True

    def execute(self):
        return max(SINK.NCC(self.x, self.y))


class NCC_Compressed(CorrelationProtocol):
    def __init__(self, x, y, **kwargs):
        super().__init__(x, y, **kwargs)
        self.similarity = True

    def execute(self):
        return max(SINK.NCC(self.x, self.y, **self.kwargs))


class SINK_protocol(CorrelationProtocol):
    def __init__(self, x, y, **kwargs):
        super().__init__(x, y, **kwargs)
        self.similarity = True

    def execute(self):
        # check if compressed here
        return SINK.SINK(self.x, self.y, **self.kwargs)


class SINK_compressed(CorrelationProtocol):
    def __init__(self, x, y, **kwargs):
        super().__init__(x, y, **kwargs)
        self.similarity = True

    def execute(self):
        return SINK.SINK(self.x, self.y, **self.kwargs)


correlation_protocols = {
    "Pearson": Pearson,
    "ED": ED,
    "NCC": NCC,
    "NCC_compressed": NCC_Compressed,
    "SINK": SINK_protocol,
    "SINK_compressed": SINK_compressed
}



#!/usr/bin/python
from numpy import zeros, sqrt, dot, diag
from numpy.linalg import svd, LinAlgError
from scipy.linalg import svd as scipy_svd


class FrequentDirections:
    def __init__(self, d, ell):
        self.d = d
        self.ell = ell
        self.m = 2 * self.ell
        self._sketch = zeros((self.m, self.d))
        self.nextZeroRow = 0

    def append(self, vector):
        if self.nextZeroRow >= self.m:
            self.__rotate__()
        self._sketch[self.nextZeroRow, :] = vector
        self.nextZeroRow += 1

    def __rotate__(self):
        try:
            [_, s, Vt] = svd(self._sketch, full_matrices=False)
        except LinAlgError as err:
            [_, s, Vt] = scipy_svd(self._sketch, full_matrices=False)

        if len(s) >= self.ell:
            sShrunk = sqrt(s[:self.ell] ** 2 - s[self.ell - 1] ** 2)
            self._sketch[:self.ell:, :] = dot(diag(sShrunk), Vt[:self.ell, :])
            self._sketch[self.ell:, :] = 0
            self.nextZeroRow = self.ell
        else:
            self._sketch[:len(s), :] = dot(diag(s), Vt[:len(s), :])
            self._sketch[len(s):, :] = 0
            self.nextZeroRow = len(s)

    def get(self):
        return self._sketch[:self.ell, :]


if __name__ == '__main__':
    import sys
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('-d', type=int, required=True, help='dimension of row vectors (number of columns in matrix).')
    parser.add_argument('-ell', type=int, required=True, help='the number of rows the sketch can keep.')
    args = parser.parse_args()

    fd = FrequentDirections(args.d, args.ell)
    for line in sys.stdin:
        try:
            row = [float(s) for s in line.strip('\n\r').split(',')]
            assert (len(row) == args.d)
        except:
            continue
        fd.append(row)

    for row in fd.get():
        sys.stdout.write('%s\n' % (','.join('%.2E' % x for x in row.flatten())))



from GRAIL.GRAIL_core import GRAIL_rep, GRAIL_general
import numpy as np
import GRAIL.kernels as kernels
from GRAIL.exceptions import KernelNotFound

class GRAIL_general:
    def __init__(self, kernel = "SINK", d = 100, f = 0.99, r = 20, kernel_param_range = [*range(1,21)],
                 eigenvecMatrix = None, inVa = None,
                 kshape_initialization_method = "partition", dictionary_sampling = "random", **kwargs):
        if kernel == 'gak':
            self.kernel = kernels.gak
            self.kwargs = kwargs
        elif kernel == 'rbf':
            self.kernel = kernels.rbf
            self.kwargs = kwargs
        elif kernel == "kdtw":
            self.kernel = kernels.kdtw
            self.kwargs = kwargs
        else:
            raise KernelNotFound
        self.d = d
        self.f = f
        self.r = r
        self.kernel_param_range = kernel_param_range
        self.eigenvecMatrix = eigenvecMatrix
        self.inVa = inVa
        self.initialization_method = kshape_initialization_method
        self.dictionary_sampling = dictionary_sampling

    def get_representation(self, X):
        """
        Get the representation of matrix X
        :param X:
        :return:
        """
        if self.d > X.shape[0]:
            raise ValueError("The number of landmark series should be smaller than the number of time series.")
        Z_k, Zexact, self.best_gamma = GRAIL_general(X, self.kernel, self.d, self.f, self.r, self.kernel_param_range,
                                    self.eigenvecMatrix, self.inVa,
                                    self.initialization_method, self.dictionary_sampling, **self.kwargs)
        return Z_k

    def get_exact_representation(self, X):
        """
        Get the representation of matrix X
        :param X:
        :return:
        """
        if self.d > X.shape[0]:
            raise ValueError("The number of landmark series should be smaller than the number of time series.")
        Z_k, Zexact, self.best_gamma = GRAIL_general(X, self.kernel, self.d, self.f, self.r, self.GV, self.eigenvecMatrix, self.inVa,
                  self.initialization_method, self.dictionary_sampling, **self.kwargs)
        return Zexact

    def get_rep_train_test(self, TRAIN, TEST, exact = True):
        """
        Get Grail representation for TRAIN and TEST sets
        :param TRAIN:
        :param TEST:
        :return:
        """
        together = np.vstack((TRAIN, TEST))
        if exact:
              rep_together = self.get_exact_representation(together)
        else:
            rep_together = self.get_representation(together)
        rep_TRAIN = rep_together[0:TRAIN.shape[0], :]
        rep_TEST = rep_together[TRAIN.shape[0]:, :]
        return rep_TRAIN, rep_TEST





class GRAIL:

    def __init__(self, kernel = "SINK", d = 100, f = 0.99, r = 20, GV = [*range(1,21)],
                 fourier_coeff = -1, e = -1, eigenvecMatrix = None, inVa = None, gamma = None, sigma = None, initialization_method = "partition"):
        self.kernel = kernel
        self.d = d
        self.f = f
        self.r = r
        self.GV = GV
        self.fourier_coeff = fourier_coeff
        self.e = e
        self.eigenvecMatrix = eigenvecMatrix
        self.inVa = inVa
        self.initialization_method = initialization_method
        self.gamma = gamma
        self.sigma = sigma


    def get_representation(self, X):
        """
        Get the representation of matrix X
        :param X:
        :return:
        """
        if self.d > X.shape[0]:
            raise ValueError("The number of landmark series should be smaller than the number of time series.")
        if self.kernel == "SINK":
            Z_k, Zexact, self.best_gamma = GRAIL_rep(X, self.d, self.f, self.r, self.GV,
                                    self.fourier_coeff, self.e, self.eigenvecMatrix, self.inVa, self.gamma, self.initialization_method)
        elif self.kernel == "kdtw":
            pass
            #Z_k, Zexact, self.best_gamma = GRAIL_rep_kdtw(X, self.d, self.f, self.r, self.GV,
             #                                             self.sigma, self.eigenvecMatrix, self.inVa)
        return Z_k

    def get_exact_representation(self, X):
        """
        Get the representation of matrix X
        :param X:
        :return:
        """
        if self.d > X.shape[0]:
            raise ValueError("The number of landmark series should be smaller than the number of time series.")
        if self.kernel == "SINK":
            Z_k, Zexact, self.best_gamma = GRAIL_rep(X, self.d, self.f, self.r, self.GV, self.fourier_coeff, self.e, self.eigenvecMatrix, self.inVa,
                  self.gamma, self.initialization_method)
        elif self.kernel == "kdtw":
            #Z_k, Zexact, self.best_gamma = GRAIL_rep_kdtw(X, self.d, self.f, self.r, self.GV, self.sigma, self.eigenvecMatrix, self.inVa)
            pass
        return Zexact

    def get_rep_train_test(self, TRAIN, TEST, exact = True):
        """
        Get Grail representation for TRAIN and TEST sets
        :param TRAIN:
        :param TEST:
        :return:
        """
        together = np.vstack((TRAIN, TEST))
        if exact:
            rep_together = self.get_exact_representation(together)
        else:
            rep_together = self.get_representation(together)
        rep_TRAIN = rep_together[0:TRAIN.shape[0], :]
        rep_TEST = rep_together[TRAIN.shape[0]:, :]
        return rep_TRAIN, rep_TEST




import numpy as np
from tslearn.metrics import gak

def rbf(x, y, gamma):
    K = np.linalg.norm(x, y) ** 2
    K *= -gamma
    np.exp(K, K)  # exponentiate K in-place
    return K

def gak(s1, s2, sigma = 1.0):
    return gak(s1, s2, sigma)

# Python source code for the "Kernelized" Dynamic Time Warping similarity (as defined in the reference below).
# Author: Pierre-Francois Marteau
# Version: V1.0 du 13/09/2014,
# Licence: GPL
# ******************************************************************
# This software and description is free delivered "AS IS" with no
# guaranties for work at all. Its up to you testing it modify it as
# you like, but no help could be expected from me due to lag of time
# at the moment. I will answer short relevant questions and help as
# my time allow it. I have tested it played with it and found no
# problems in stability or malfunctions so far.
# Have fun.
# *****************************************************************
# Please cite as:
# @article{marteau:hal-00486916,
#   AUTHOR = {Marteau, Pierre-Francois and Gibet, Sylvie},
#   TITLE = {{On Recursive Edit Distance Kernels with Application to Time Series Classification}},
#   JOURNAL = {{IEEE Transactions on Neural Networks and Learning Systems}},
#   PAGES = {1-14},
#   YEAR = {2014},
#   MONTH = Jun,
#   KEYWORDS = {Elastic distance, Time warp kernel, Time warp inner product, Definiteness, Time series classification, SVM},
#   DOI = {10.1109/TNNLS.2014.2333876},
#   URL = {http://hal.inria.fr/hal-00486916}
# }
#
''''
# function kdtw(A, B, sigma, epsilon=1e-3)
# Dynamic programming implementation of KDTW kernel
# input A: first multivariate time series: array of array (nxd), n is the number of sample, d is the dimension of each sample
# intput B: second multivariate time series: array of array (nxd), n is the number of sample, d is the dimension of each sample
# input sigma: >0 used in the exponential local kernel 
# output similarity: similarity between A and B (the higher, the more similar)
'''


def kdtw(A, B, sigma=1, epsilon=1e-3):
    d = np.shape(A)[1]
    Z = [np.zeros(d)]
    A = np.concatenate((Z, A), axis=0)
    B = np.concatenate((Z, B), axis=0)
    [la, d] = np.shape(A)
    [lb, d] = np.shape(B)

    DP = np.zeros((la, lb))
    DP1 = np.zeros((la, lb));
    DP2 = np.zeros(max(la, lb));
    l = min(la, lb);
    DP2[1] = 1.0;
    for i in range(1, l):
        DP2[i] = Dlpr(A[i], B[i], sigma, epsilon);
    if la < lb:
        for i in range(la, lb):
            DP2[i] = Dlpr(A[la - 1], B[i], sigma, epsilon);
    elif lb < la:
        for i in range(lb, la):
            DP2[i] = Dlpr(A[i], B[lb - i], sigma, epsilon);

    DP[0, 0] = 1;
    DP1[0, 0] = 1;
    n = len(A);
    m = len(B);

    for i in range(1, n):
        DP[i, 1] = DP[i - 1, 1] * Dlpr(A[i], B[2], sigma, epsilon);
        DP1[i, 1] = DP1[i - 1, 1] * DP2[i];

    for j in range(1, m):
        DP[1, j] = DP[1, j - 1] * Dlpr(A[2], B[j], sigma, epsilon);
        DP1[1, j] = DP1[1, j - 1] * DP2[j];

    for i in range(1, n):
        for j in range(1, m):
            lcost = Dlpr(A[i], B[j], sigma, epsilon);
            DP[i, j] = (DP[i - 1, j] + DP[i, j - 1] + DP[i - 1, j - 1]) * lcost;
            if i == j:
                DP1[i, j] = DP1[i - 1, j - 1] * lcost + DP1[i - 1, j] * DP2[i] + DP1[i, j - 1] * DP2[j]
            else:
                DP1[i, j] = DP1[i - 1, j] * DP2[i] + DP1[i, j - 1] * DP2[j];
    DP = DP + DP1;
    return DP[n - 1, m - 1]


# local similarity between two samples
# a: 1d numpy array
# b: 1d numpy array
# input sigma: >0 used in the exponential local kernel
# input epsilon: 1 > epsilon > 0
# return the local matching similarity (probability)
def Dlpr(a, b, sigma=1, epsilon=1e-3):
    return (np.exp(-np.sum((a - b) ** 2) / sigma) + epsilon) / (3 * (1 + epsilon))




import numpy as np
import GRAIL.Correlation_protocol as Correlation_protocol
import GRAIL.exceptions as exceptions


class Correlation:

    def __init__(self, x, y, correlation_protocol_name, time_window=None, **kwargs):
        """

        :param x: time series. Can be either a path to a csv file or list
        :param y: time series. Can be either a path to a csv file or list
        :param correlation_protocol_name:
        :param time_window: the time window in which the series are being correlated.
        :param args: arguments for the correlation algorithm.
        """

        self.x = self.load(x)
        self.y = self.load(y)
        self.correlation_protocol_name = correlation_protocol_name
        try:
            self.protocol_class = Correlation_protocol.correlation_protocols[self.correlation_protocol_name]
        except KeyError:
            raise exceptions.ProtocolNotFound("Correlation Protocol" + correlation_protocol_name + "not found")

        if kwargs:
            self.protocol = self.protocol_class(x, y, **kwargs)
        else:
            self.protocol = self.protocol_class(x,y)

        self.check_entries()
        if time_window:
            self.time_window = time_window
            x = self.crop(x, time_window)
            y = self.crop(y, time_window)

    def is_similarity(self):
        return self.protocol.similarity

    @classmethod
    # Be careful about this. Will not work later.
    def is_similarity(cls, correlation_protocol_name):
        if correlation_protocol_name == "ED":
            return False
        else:
            return True

    def check_entries(self):
        if hasattr(self, "time_window"):
            if (not isinstance(self.time_window, list)) or (len(self.time_window) != 2):
                raise TypeError("time window should be a list of 2 elements.")

    def load(self, x):
        if isinstance(x, list):
            return np.array(x)
        elif isinstance(x, np.generic) or isinstance(x, np.ndarray):
            pass
        else:
            raise TypeError("Time series should be a list, or a numpy array")

    def crop(self, x, window):
        return x[window[0]:window[1]]

    def correlate(self):
        return self.protocol.execute()

    def is_correlated(self, threshold):
        return self.protocol.execute() >= threshold


